/**
 * @file tuya_app_main.c
 * @author www.tuya.com
 * @brief tuya_app_main module is used to 
 * @version 0.1
 * @date 2022-10-28
 *
 * @copyright Copyright (c) tuya.inc 2022
 *
 */

#include "tuya_cloud_types.h"
#include "tuya_iot_wifi_api.h"
#include "tuya_ws_db.h"
#include "gw_intf.h"
#include "tal_log.h"
#include "tal_wifi.h"
#include "tal_uart.h"

#include "tuya_ringbuf.h"

#if defined(ENABLE_LWIP) && (ENABLE_LWIP == 1)
#include "lwip_init.h"
#endif

/***********************************************************
************************macro define************************
***********************************************************/

#define PID "uiiyltkjmkhwumga"

/***********************************************************
***********************typedef define***********************
***********************************************************/

#pragma pack(1)
typedef struct data_frame{
    uint16_t frame_head;
    uint8_t data_len;
    uint8_t data[0];
};
#pragma pack()

/***********************************************************
********************function declaration********************
***********************************************************/


/***********************************************************
***********************variable define**********************
***********************************************************/
/* app thread handle */
STATIC THREAD_HANDLE ty_app_thread = NULL;

/***********************************************************
***********************function define**********************
***********************************************************/

STATIC VOID_T user_main(VOID_T)
{
    OPERATE_RET rt = OPRT_OK;

    /* Initialization, because DB initialization takes a long time, 
     * which affects the startup efficiency of some devices, 
     * so special processing is performed during initialization to delay initialization of DB
     */
    TY_INIT_PARAMS_S init_param = {0};
    init_param.init_db = TRUE;
    strcpy(init_param.sys_env, TARGET_PLATFORM);
    TUYA_CALL_ERR_LOG(tuya_iot_init_params(NULL, &init_param));

    tal_log_set_manage_attr(TAL_LOG_LEVEL_DEBUG);

    TAL_UART_CFG_T cfg;
    memset(&cfg, 0, sizeof(TAL_UART_CFG_T));
    cfg.base_cfg.baudrate = 115200;
    cfg.base_cfg.databits = TUYA_UART_DATA_LEN_8BIT;
    cfg.base_cfg.parity = TUYA_UART_PARITY_TYPE_NONE;
    cfg.base_cfg.stopbits = TUYA_UART_STOP_LEN_1BIT;
    cfg.rx_buffer_size = 1024;

    tal_uart_init(TUYA_UART_NUM_0, &cfg);

    uint8_t tmp_buf[255] = {0};
    uint8_t tmp_buf_2[255] = {0};
    uint16_t read_len = 0;

    // 使用 ringbuffer 来缓存接收到的数据
    TUYA_RINGBUFF_T recv_cache;
    uint16_t cache_len = 0;
    TUYA_CALL_ERR_LOG(tuya_ring_buff_create(255, OVERFLOW_STOP_TYPE, &recv_cache));

    int i = 0;

    for (;;) {
        memset(tmp_buf, 0, 255);
        read_len = tal_uart_read(TUYA_UART_NUM_0, tmp_buf, 255);
        if (read_len > 0) {
            TAL_PR_DEBUG("recv data, len: %d", read_len);
            // 缓存本次接收到的数据
            tuya_ring_buff_write(recv_cache, tmp_buf, read_len);
            // 获取已经缓存数据的长度
            cache_len = tuya_ring_buff_used_size_get(recv_cache);
            if (cache_len > 3) {
                tuya_ring_buff_peek(recv_cache, tmp_buf, cache_len);
                for (i=0; i<cache_len; i++) {
                    if (tmp_buf[i] == 0x55 && tmp_buf[i+1] == 0xAA) { //找到帧头
                        if(i>0) { // 丢弃帧头之前的无效数据
                            tuya_ring_buff_read(recv_cache, tmp_buf_2, i);
                        }
                        break;
                    }
                }
                // 如果没有找到帧头，丢弃所有缓存数据
                if (i == cache_len) {
                    TAL_PR_DEBUG("discard all");
                    tuya_ring_buff_read(recv_cache, tmp_buf_2, cache_len);
                    continue;
                }

                // 判断接收到的一帧是否接收完整
                cache_len = tuya_ring_buff_used_size_get(recv_cache);
                if (cache_len > 3) {
                    tuya_ring_buff_peek(recv_cache, tmp_buf, cache_len);
                    struct data_frame *recv_frame = (struct data_frame *)tmp_buf;
                    if (recv_frame->data_len + 3 <= cache_len) {
                        // 接收到完整的一帧，读出
                        TAL_PR_DEBUG("get one frame");
                        uint8_t frame_len = recv_frame->data_len + 3;
                        tuya_ring_buff_read(recv_cache, tmp_buf, frame_len);
                        TAL_PR_HEXDUMP_DEBUG("frame", tmp_buf, frame_len);
                        tal_uart_write(TUYA_UART_NUM_0, tmp_buf, frame_len);
                    }
                }
                continue;
            }
        }
    }

    return;
}

/**
* @brief  task thread
*
* @param[in] arg:Parameters when creating a task
* @return none
*/
STATIC VOID_T tuya_app_thread(VOID_T *arg)
{
    /* Initialization LWIP first!!! */
#if defined(ENABLE_LWIP) && (ENABLE_LWIP == 1)
    TUYA_LwIP_Init();
#endif

    user_main();

    tal_thread_delete(ty_app_thread);
    ty_app_thread = NULL;
}

/**
 * @brief user entry function
 *
 * @param[in] none: 
 *
 * @return none
 */
VOID_T tuya_app_main(VOID_T)
{
    THREAD_CFG_T thrd_param = {4096, 4, "tuya_app_main"};
    tal_thread_create_and_start(&ty_app_thread, NULL, NULL, tuya_app_thread, NULL, &thrd_param);
}
